(* Source file : typing/typecheck.sml, Copyright 1989 by AT&T Bell Laboratories 

 Modified for typechecking with polymorphic recursion (Milner-Mycroft-types):
 by M.Emms, June 1995/March 1997 and H.Leiss,April 1997/Oct 1997/March 1998
 
 (Copyright (c) M.Emms,H.Leiss 1997, {emms,leiss}@cis.uni-muenchen.de)         

 *)

signature TYPECHECK = 
    sig

        val decType : StaticEnv.staticEnv * Absyn.dec * bool
            * ErrorMsg.errorFn * SourceMap.region -> Absyn.dec
        val debugging : bool ref

    end (* signature TYPECHECK *)


structure Typecheck : TYPECHECK =
    struct

        local open Array List Types VarCon BasicTypes TypesUtil Unify Absyn
            Overload ErrorMsg PrettyPrint PPUtil PPType PPAbsyn

              structure SE = StaticEnv
              structure II = InlInfo
              structure DA = Access
              structure EU = ElabUtil
              structure ED = ElabDebug
              structure OLL = OverloadLit
              structure PPTL = PPTypelist
                  
        in 

            (* debugging *)
            val say = Control.Print.say
            val debugging = Control.debugging (* ref false *)
            fun debugmsg (msg: string) = if !debugging then (say msg; say "\n") else ()
            val debugPrint = (fn x => ED.debugPrint debugging x)

            fun bug msg = ErrorMsg.impossible("TypeCheck: "^msg)


            infix 9 sub
            infix -->

            val printDepth = Control.Print.printDepth

            fun refNewDcon(DATACON{name,const,rep,typ,sign}) = 
                DATACON{name=name,const=const,rep=rep,typ=refPatType,sign=sign}

            exception NotThere
            exception Subtract

            fun message(msg,mode: Unify.unifyFail) =
                String.concat[msg," [",Unify.failMessage mode,"]"]

            fun mkDummy0 () = BasicTypes.unitTy

            (*
             * decType : SE.staticEnv * A.dec * bool * EM.errorFn * region -> A.dec 
             *)
            fun decType(env,dec,toplev,err,region) = 
                let

                    val ppType = PPType.ppType env
                    val ppPat = PPAbsyn.ppPat env
                    val ppExp = PPAbsyn.ppExp(env,NONE)
                    val ppRule = PPAbsyn.ppRule(env,NONE)
                    val ppVB = PPAbsyn.ppVB(env,NONE)
                    val ppRVB = PPAbsyn.ppRVB(env,NONE)
                    val ppDec = (fn ppstrm => fn d => PPAbsyn.ppDec (env,NONE) ppstrm (d,!printDepth))

                    fun debugPPDec (msg,dec) =
                        ED.withInternals(fn () => ED.debugPrint debugging (msg, ppDec, dec))

                    fun debugPPType (msg,ty) =
                        ED.withInternals(fn () => ED.debugPrint debugging (msg, ppType, ty))

                    fun debugPPTyvar tv = 
                        ED.withInternals(fn () => debugmsg (PPType.tyvarPrintname tv))


                    fun unifyErr{ty1,name1,ty2,name2,message=m,region,kind,kindname,phrase} =
                        (unifyTy(ty1,ty2); true) handle Unify(mode) =>
                            (err region COMPLAIN (message(m,mode))
                             (fn ppstrm => 
                              (PPType.resetPPType();
                               let val len1= size name1 
                                   val len2= size name2
                                   val spaces = "                                   "
                                   val pad1= substring(spaces,0,Int.max(0,len2-len1))
                                   val pad2= substring(spaces,0,Int.max(0,len2-len1))
                                   val m = if m="" then name1 ^ " and " ^ name2 ^ " don't agree"
                                           else m
                               in if name1="" then ()
                                  else (add_newline ppstrm; add_string ppstrm (name1 ^ ": " ^ pad1);
                                        ppType ppstrm ty1); 
                                      if name2="" then ()
                                      else (add_newline ppstrm; add_string ppstrm (name2 ^ ": " ^ pad2);
                                            ppType ppstrm ty2);
                                          if kindname="" then ()
                                          else (add_newline ppstrm; add_string ppstrm("in "^kindname^":");
                                                add_break ppstrm (1,2); kind ppstrm (phrase,!printDepth))
                               end));
                             false)

                    val _ = debugmsg (">>decType: toplev = " ^ Bool.toString toplev)
                    val _ = debugPPDec(">>decType: dec = ",dec)

                    (* Printing a semiunification problem *)

                    val ppTypelist = PPTL.pptypelist env
                    val ppIneqlist = PPTL.ppineqlist env
                    val g = PrettyPrint.with_pp(ErrorMsg.defaultConsumer())

                    fun nameprint name = (g (fn ppstream => ppSym ppstream name); print ":")
                    and typprint typ = (g (fn ppstream => ppType ppstream typ); print "\n")
                    and inpr ilist   = (g (fn ppstream => ppIneqlist ppstream ilist))
                    and lpr tlist    = (g (fn ppstream => ppTypelist ppstream tlist))
                    and rvbpr dec    = (g (fn ppstream => ppDec ppstream dec))
                    and rulepr exp   = (g (fn ppstream => ppRule ppstream (exp,!printDepth)))
                    and letpr exp    = (g (fn ppstream => ppExp ppstream (exp,!printDepth)))
                    and monopr monoConstr = 
                        let 
                            val monos = (Semiunify.monoConstraintsAsIneqns monoConstr)
                        in 
                            if length monos > 0 then 
                                (print "the monomorphic vars (instantiated):\n";
                                 inpr monos)
                            else ()
                        end

                    fun revadd (a::acc) list = revadd acc (a::list)
                      | revadd [] list = list

                    (* Global variables for semiunification: *)

                    val rulebound = ref([] : ty list)             (* types in lambda-assumptions *)
                    val recbound  = ref([] : VarCon.var list)     (* indiv.recvars during recDec *)
                    val semiprob  = ref([] : Semiunify.ineq list) (* semiunfication constraints  *)
                    val monoConstraints 
                        = ref([] : Semiunify.monoconstraint list) (* monomorphic tyvars *)
                    val topprob   = ref true
                    val eqcounter = ref 0                         (* inequation counter - var occ.no *)
                    val printSolvedIneqns = ref true

                    (* Viewing intermediate steps of the semiunification solving process *)

                    fun showIneqns (ineqnsL,ineqnsR) thestring =
                        (print (thestring^" print? ");
                         let 
                             val y = TextIO.inputLine(TextIO.stdIn)
                         in
                             if y = "y\n" 
                                 then (inpr (revadd ineqnsR ineqnsL);
                                       monopr (!monoConstraints);
                                       print "\n";
                                       (ineqnsL,ineqnsR)) 
                             else if y = "stop\n" then 
                                 (Control.CG.viewsemi := 1; (ineqnsL,ineqnsR)) 
                                  else (ineqnsL,ineqnsR)
                         end)

                    val _ = Semiunify.show := showIneqns

                    fun checkvarpr varname semiprob ineqnsByCopying = 
                        if (!Control.CG.viewsemi) > (!(# Var Control.CG.showsemi))
                            then
                                (print ("Show var case for occurrence of: " ^ varname ^ " ? ");
                                 let 
                                     val y = TextIO.inputLine(TextIO.stdIn)
                                 in (if y = "y\n" then 
                                         (print "\nnew inequations by copying:\n";
                                          inpr ineqnsByCopying;
                                          print "other inequations: \n";
                                          inpr (!semiprob); 
                                          print "\n") 
                                     else (if y = "stop\n" then 
                                               Control.CG.viewsemi := 1
                                           else ()))
                                 end)
                        else ()

                    fun checkrulepr1 therule pty semiprob =
                        if (!Control.CG.viewsemi) > (!(# Rule Control.CG.showsemi)) then
                            (print "Show rule case for: \n";
                             (rulepr therule); 
                             print " ? ";
                             let val y = TextIO.inputLine(TextIO.stdIn) 
                             in if y = "y"^"\n" then
                                 (print "\n"; print "the pattern type is: \n"; 
                                  typprint pty;
                                  print "\nthe inequations are: \n";
                                  inpr (!semiprob);
                                  monopr (!monoConstraints);
                                  print "\n";
                                  true)
                                else if y = "stop"^"\n" then 
                                    (Control.CG.viewsemi := 1; false)
                                     else false
                             end)
                        else false

                    and checkrulepr2 printAfterRule semiprob =
                        if printAfterRule then 
                            (print "The inequations after typing the rule are: \n";
                             inpr (!semiprob);
                             monopr (!monoConstraints);
                             print "\n")
                        else ()

                    fun checkletpr1 thelet semiprob = 
                        if (!Control.CG.viewsemi) > (!(# Let Control.CG.showsemi)) then
                            (print "Show let case for: \n";
                             letpr thelet;
                             print " ? ";
                             let val y = TextIO.inputLine(TextIO.stdIn) 
                             in (if y = "y\n" 
                                     then (print "Inequations before typing the let-expression: \n";
                                           inpr (!semiprob);
                                           monopr (!monoConstraints);
                                           print "\n";
                                           true)
                                 else false)
                             end)
                        else false

                    and checkletpr2 printAfterLet semiprob = 
                        if printAfterLet then 
                            (print "Inequations after typing the let-expression:\n";
                             inpr (!semiprob);
                             monopr (!monoConstraints);
                             print "\n")
                        else ()

                    fun checkrvbpr1 rvbs semiprob monoConstraints = 
                        if (!Control.CG.viewsemi) > (!(# Valrec Control.CG.showsemi)) 
                            then (print "Show recursive value case for:\n";
                                  rvbpr (VALRECdec (rvbs));
                                  print " ? ";
                                  let val y = TextIO.inputLine(TextIO.stdIn) 
                                  in (if y = "y"^"\n" then 
                                          (print "\nInequations before solving: \n";
                                           inpr (!semiprob);
                                           monopr (!monoConstraints);
                                           print "\nStep through solving the inequations? ";
                                           let val stepping = TextIO.inputLine(TextIO.stdIn)
                                           in (if stepping = "y\n" then 
                                                   (# Step Control.CG.showsemi) := 
                                                   (!(# Valrec Control.CG.showsemi)) 
                                               else 
                                                   (# Step Control.CG.showsemi) := 
                                                   (!Control.CG.viewsemi))
                                           end;
                                           true)
                                      else false) 
                                  end)
                        else false

                    and checkrvbpr2 solved semiprob monoConstraints =
                        if solved = true then 
                            (print "Inequations after solving: \n";
                             inpr (!semiprob);
                             monopr (!monoConstraints);
                             print "\n")
                        else ()


                    (* Updating the list of rulebound tyvars *)

                    fun rbupdate ty = case (prune ty) of 
                        CONty(tyc,args) => app rbupdate args
                      | VARty(ref(OPEN{kind=FLEX(fields),...}))
                            => app (rbupdate o #2) fields
                      | v as VARty(ref(OPEN{kind=META,...}))
                            => (rulebound := (v::(!rulebound)))
                      | v as (VARty(tvinfo)) (* UBOUND | LITERAL | SCHEME | LBOUND *)
                            => (rulebound := (v::(!rulebound)))
                      | _ => () 
                    and isRulebound ty = 
                        let fun there x (y::z) = if equalType(x,y) then true else there x z
                              |  there _ [] = false
                        in there ty (!rulebound) 
                        end

(* For polymorphic recursion, expType needs to know if an individual variable is 
 mono- or poly-bound. Since VALvar{typ,name,access} does not have this attribute,
 to test the property we stack the rec-bound vars and pop after recdecs).
 *) 

fun polybound v = case v of 
    VALvar{typ=ref(POLYty _),...} => true        (* applied occurrences of dec-vars *)
  | VALvar{typ=ref(VARty(ref(OPEN{kind=META,...}))),...} =>
	member v (!recbound)  (* occurrences of rec-dec vars in defining expression *)
  | _ => false 
and member x [] = false
  | member x (y::ys) = if eqVar(x,y) then true else member x ys
and eqVar(VALvar{access=accx,...},VALvar{access=accy,...}) = (accx = accy) 
  | eqVar(_,_) = false

(* To type a variable occurrence, copy the (body of) the assumed type *) 

fun copyTy(ty) =
    let val copymap = ref([]:(ty * ty) list)

        fun fetchCopy ty =
	    let fun find [] = raise NotThere
		  | find((ty1,ty2)::rest) = if equalType(ty,ty1) then ty2 
                                            else find rest
            in find(!copymap)
	    end

        fun fetchRulebound ty = if isRulebound ty then ty else raise NotThere

        fun copy(ty) = case prune(ty) 
            of VARty(tv as ref(OPEN{depth=d,eq=e,kind=k})) =>
                (case k 
                     of META => 
                         (fetchCopy(VARty(tv)) handle NotThere =>  
                             fetchRulebound(VARty(tv)) handle NotThere =>
                                 (* NOTE 1: although the rulebound list of the occurrence of 
                                  indiv.var label is bigger than monoConstraints(label,_), 
                                  ty cannot contain tyvars of rulebound-monoconstr. *)
                                 if (d=infinity orelse d=0) then VARty(tv)
                                 else let val tv' = ref(!tv)
                                      in (copymap := (VARty(tv),VARty(tv'))::(!copymap);
                                          VARty(tv'))
                                      end)                           
                   | FLEX(fields) =>
                             let fun f (lab,ty) = (lab,copy ty) 
                             in (tv := OPEN{kind=FLEX(map f fields),depth=d,eq=e};
                                 VARty(tv))
                             end)
          | VARty(tv as ref(UBOUND{...})) => VARty(tv)
          | CONty(tyc,args) => CONty(tyc, map copy args) 
          | POLYty{sign=polysign,tyfun=tyfun} => POLYty{sign=polysign,tyfun=tyfun} (* ??? *)
          | IBOUND(n) => IBOUND(n)
          | WILDCARDty => WILDCARDty
          | UNDEFty => UNDEFty
          | tp => tp  (* '93: VARty(tv as ref(UBOUND(_)) => VARty(tv) *)
    in 
        (copy(ty),!copymap)
    end

fun copy label ty =  
    let 
        val _ = Semiunify.compress rulebound   
        (* NOTE 2: need only compress monoconstraints(label), but see NOTE 1 *)
        val (ty',pairs) = copyTy(ty)
        val _ = case pairs of nil => () | (p::ps) => (eqcounter := (!eqcounter) + 1)
        val ineqnsByCopying = 
            (map (fn (x,y) => Semiunify.ineq(label,!eqcounter,x,y)) pairs)
        val (Semiunify.mylab(_,var)) = label
        val _ = checkvarpr (Symbol.name var) semiprob ineqnsByCopying 
        val _ = semiprob := revadd ineqnsByCopying (!semiprob)
    in
        ty'
    end

val _ = debugmsg (">>decType: toplev = " ^ Bool.toString toplev)
val _ = debugPPDec(">>decType: dec = ",dec)

fun generalizeTy(VALvar{typ,path,...}, userbound: tyvar list,
		 occ:occ, generalize: bool, region) : tyvar list =
    let val _ = debugmsg ("generalizeTy: "^SymPath.toString path)
	val _ = debugmsg ("userbound: ")
	val _ = List.app debugPPTyvar userbound

	val failure = ref false
	val mkDummy = if toplevel occ
	              then TypesUtil.dummyTyGen()
		      else mkDummy0 (* shouldn't be called *)

	val index = ref 0  (* counts no of type variables bound *)
	fun next() = let val i = !index in index := i+1; i end
	val sign = ref([]: Types.polysign)
	fun localUbound tv =
	    let fun mem(tv'::rest) = eqTyvar(tv,tv') orelse mem rest
		  | mem [] = false
	     in mem userbound
	    end
	(* menv: a reference to an association list environment mapping
	 *   generalized tyvars to the corresponding IBOUND type. 
	 * ASSERT: there are no duplicate tyvars in domain of menv. *)
	val menv = ref([]: (tyvar*ty) list)
	fun lookup tv =
	    let fun find [] = raise NotThere
		  | find((tv',ty)::rest) = if eqTyvar(tv,tv') then ty 
							      else find rest
	     in find(!menv)
	    end
	fun bind(tv,ty) = menv := (tv,ty) :: !menv
	fun gen(ty) =     
	    case ty
	     of VARty(ref(INSTANTIATED ty)) => gen ty
	      | VARty(tv as ref(OPEN{depth,eq,kind})) =>
		  (case kind
		     of FLEX[(lab,_)] =>
                         if ((depth > lamdepth occ) andalso
                             (generalize orelse (toplevel occ)))
                            orelse ((toplevel occ) andalso (depth=0))
                         then
			   (err region COMPLAIN (String.concat
			     ["unresolved flex record\n\
			      \   (can't tell what fields there are besides #",
			      Symbol.name lab, ")"])
			    nullErrorBody;
			   WILDCARDty)
                         else ty
		      | FLEX _ =>
                         if ((depth > lamdepth occ) andalso
                             (generalize orelse (toplevel occ)))
                            orelse ((toplevel occ) andalso (depth=0))
                         then
  			   (err region COMPLAIN 
			        "unresolved flex record (need to know the \
			        \names of ALL the fields\n in this context)"
			    (fn ppstrm =>
			       (PPType.resetPPType();
				add_newline ppstrm;
				add_string ppstrm "type: ";
				ppType ppstrm ty));
			    WILDCARDty)
                         else ty
		      | META =>
			  if depth > lamdepth occ
			  then if generalize then
				  lookup tv handle NotThere =>
				    let val new = IBOUND(next())
				     in sign := eq :: !sign;
				        bind(tv,new); 
					new
				    end
			       else (if toplevel occ
				     then let val new = mkDummy()
					   in failure := true;
                                              tv := INSTANTIATED new;
					      new
					  end
				     else (if !Control.valueRestrictionLocalWarn
					   then err region WARN
				             ("type variable not generalized\
                                              \ in local decl (value restriction): "
                                              ^ (tyvarPrintname tv))
				             nullErrorBody
					   else ();
					   (* reset depth to prevent later
					      incorrect generalization inside
					      a lambda expression.  See typechecking
					      test 5.sml *)
					   tv := OPEN{depth = lamdepth occ,
						      eq = eq, kind = kind};
					   ty))
			  else if toplevel occ andalso depth = 0
			   (* ASSERT: failed generalization at depth 0.
			      see bug 1066. *)
			    then lookup tv handle NotThere =>
				 let val new = mkDummy()
				  in failure := true;
                                     tv := INSTANTIATED new;
				     new
				 end
			  else ty) (* raise SHARE *)
	      | VARty(tv as ref(UBOUND{name,depth,eq})) =>
		 (debugmsg ("UBOUND:" ^Symbol.name name);
		  if localUbound tv
		  then (debugmsg "is local";
		       if depth > lamdepth occ andalso generalize
		       then (debugmsg "is generalized";
			     lookup tv handle NotThere =>
			      let val new = IBOUND(next())
			       in sign := eq :: !sign;
				  bind(tv,new);
				  new
			      end)
		       else (err region COMPLAIN
			     ("explicit type variable cannot be \
			       \generalized at its binding \
			       \declaration: " ^
			       (tyvarPrintname tv))
			      nullErrorBody;
			     tv := INSTANTIATED WILDCARDty;
			     WILDCARDty))
		  else (debugmsg "is not local"; ty))
	      | (VARty(ref(LITERAL _)) | VARty(ref(SCHEME _))) => ty
	      | CONty(tyc,args) => CONty(tyc, map gen args) (*shareMap*)
	      | WILDCARDty => WILDCARDty
	      | _ => bug "generalizeTy -- bad arg"

	val _ = debugPPType (">>gen: before: ",!typ)
	val ty = gen(!typ)
	val _ = debugPPType (">>gen: after: ",ty)

        val generalizedTyvars = map #1 (rev(!menv))

	fun elimUbound(tv as ref(UBOUND{depth,eq,...})) = 
              (tv := OPEN{depth=depth,eq=eq,kind=META})
          | elimUbound _ = ()
        val _ = app elimUbound generalizedTyvars

     in if !failure andalso !Control.valueRestrictionTopWarn
	  then err region WARN
	        "type vars not generalized because of\n\
                 \   value restriction are instantiated to dummy types (X1,X2,...)"
		nullErrorBody
          else ();

	typ := POLYty{sign = rev(!sign),
		      tyfun = TYFUN{arity=(!index),body=ty}};

        generalizedTyvars  (* return the tyvars that were generalized *)
    end

  | generalizeTy _ = bug "generlizeTy - bad arg"
  

fun generalizePat(pat: pat, userbound: tyvar list, occ: occ, 
                  generalize: bool, region) =
    let val tvs : tyvar list ref = ref []
        fun gen(VARpat v) = 
	      (let val x = generalizeTy(v,userbound,occ,generalize,region)
                   val _ = case (x, !tvs) 
                            of (_::_, _::_) => bug "generalizePat 1234"
                             | _ => ()
                in tvs := (x@(!tvs))
               end)
	  | gen(RECORDpat{fields,...}) = app (gen o #2) fields
	  | gen(APPpat(_,_,arg)) = gen arg
	  | gen(CONSTRAINTpat(pat,_)) = gen pat
	  | gen(LAYEREDpat(varPat,pat)) = (gen varPat; gen pat)
	  | gen _ = ()
     in gen pat; !tvs
    end

fun applyType(ratorTy: ty, randTy: ty) : ty =
  let val resultType = mkMETAty()
   in unifyTy(ratorTy, (randTy --> resultType)); resultType
  end

fun patType(pat: pat, depth, region) : pat * ty =
    case pat
      of WILDpat => (pat,mkMETAtyBounded depth)
       | VARpat(VALvar{typ as ref UNDEFty,...}) => 
	      (typ := mkMETAtyBounded depth; (pat,!typ))
			             (* multiple occurrence due to or-pat *)
       | VARpat(VALvar{typ, ...}) => (pat, !typ) 
       | INTpat (_,ty) => (OLL.push ty; (pat,ty))
       | WORDpat (_,ty) => (OLL.push ty; (pat,ty))
       | REALpat _ => (pat,realTy)
       | STRINGpat _ => (pat,stringTy)
       | CHARpat _ => (pat,charTy)
       | RECORDpat{fields,flex,typ} =>
	   (* fields assumed already sorted by label *)
	   let fun g(lab,pat') = 
                 let val (npat,nty) = patType(pat',depth,region)
                  in ((lab,npat), (lab,nty))
                 end
               val (fields',labtys) = mapUnZip g fields
               val npat = RECORDpat{fields=fields',flex=flex,typ=typ}
	    in if flex
	       then let val ty = VARty(mkTyvar(mkFLEX(labtys,depth)))
		     in typ := ty; (npat,ty)
		    end
	       else (npat,recordTy(labtys))
	   end
       | VECTORpat(pats,_) => 
          (let val (npats,ntys) = 
                     mapUnZip (fn pat => patType(pat,depth,region)) pats
               val nty =
	       foldr (fn (a,b) => (unifyTy(a,b); b)) (mkMETAtyBounded depth) ntys
            in (VECTORpat(npats,nty), CONty(vectorTycon,[nty]))
           end handle Unify(mode) => (
	     err region COMPLAIN 
		 (message("vector pattern type failure",mode)) nullErrorBody;
	     (pat,WILDCARDty)))
       | ORpat(p1, p2) => 
           let val (p1, ty1) = patType(p1, depth, region)
  	       val (p2, ty2) = patType(p2, depth, region)
	   in
	     unifyErr{ty1=ty1,ty2=ty2,name1="expected",name2="found",
		      message="or-patterns don't agree",region=region,
		      kind=ppPat,kindname="pattern",phrase=pat};
	     (ORpat(p1, p2), ty1)
	   end
       | CONpat(dcon as DATACON{typ,...},_) => 
           let val (ty, insts) = instantiatePoly typ
               (* the following is to set the correct depth information
                * to the type variables in ty. (ZHONG)
                *)
               val nty = mkMETAtyBounded depth
               val _ = unifyTy(nty, ty) 
            in (CONpat(dcon,insts),ty)
           end
       | APPpat(dcon as DATACON{typ,rep,...},_,arg) =>
	   let val (argpat,argty) = patType(arg,depth,region)
               val (ty1,ndcon) = case rep
                                  of DA.REF => (refPatType,refNewDcon dcon)
                                   | _ => (typ,dcon)
               val (ty2,insts) = instantiatePoly ty1
               val npat = APPpat(ndcon,insts,argpat)
            in (npat,applyType(ty2,argty))
	       handle Unify(mode) =>
		(err region COMPLAIN
                  (message("constructor and argument don't agree in pattern",mode))
		  (fn ppstrm =>
		   (PPType.resetPPType();
		    add_newline ppstrm;
		    add_string ppstrm "constructor: ";
		    ppType ppstrm typ; add_newline ppstrm;
		    add_string ppstrm "argument:    ";
		    ppType ppstrm argty; add_newline ppstrm;
		    add_string ppstrm "in pattern:"; add_break ppstrm (1,2);
		    ppPat ppstrm (pat,!printDepth)));
		 (pat,WILDCARDty))
	   end
       | CONSTRAINTpat(pat',ty) => 
	   let val (npat,patTy) = patType(pat',depth,region)
	    in if unifyErr{ty1=patTy,name1="pattern",ty2=ty,name2="constraint",
			message="pattern and constraint don't agree",
			region=region,kind=ppPat,kindname="pattern",phrase=pat}
		then (CONSTRAINTpat(npat,ty),ty)
		else (pat,WILDCARDty)
	   end
       | LAYEREDpat(vpat as VARpat(VALvar{typ,...}),pat') =>
           let val (npat,patTy) = patType(pat',depth,region)
               val _ = (typ := patTy)
            in (LAYEREDpat(vpat,npat),patTy)
           end
       | LAYEREDpat(cpat as CONSTRAINTpat(VARpat(VALvar{typ,...}),ty),pat') =>
	   let val (npat,patTy) = patType(pat',depth,region)
	    in if unifyErr{ty1=patTy,name1="pattern",ty2=ty,name2="constraint",
			   message="pattern and constraint don't agree",
			   region=region,kind=ppPat,kindname="pattern",phrase=pat}
		   then (typ := ty; (LAYEREDpat(cpat,npat),ty))
		  else (pat,WILDCARDty)
	   end
       | p => bug "patType -- unexpected pattern"


fun expType(exp: exp, occ: occ, region) : exp * ty =
     case exp
      of VARexp(r as ref(VALvar{typ, info=II.INL_PRIM(_,SOME st), ...}), _) => 
           let val (sty, insts) = instantiatePoly(st)
               val (nty, _) = instantiatePoly(!typ)
            in unifyTy(sty, nty) handle _ => ();  (* ??? *)
               (VARexp(r, insts), sty)            
           end
        
       | VARexp(r as ref(VALvar{typ, ...}),_) => 
           let 
               val (Semiunify.mylab(_,lab)) = Findvars.find_var_var(!r)
               val varname = Symbol.name lab
               val (ty1,insts) = instantiatePoly(!typ)
               val ty = if (!Control.CG.oldchecker) then ty1 
                        else if polybound (!r) 
                                 then 
                                       (copy (Findvars.find_var_var(!r)) ty1)
                             else ty1 
	    in (VARexp(r, insts),ty)  (* should be insts[copied tyvars] ? *)
	   end
       | VARexp(refvar as ref(OVLDvar _),_) =>
 	    (exp,pushOverloaded(refvar, err region))
       | VARexp(r as ref ERRORvar, _) => (exp, WILDCARDty)
       | CONexp(dcon as DATACON{typ,...},_) => 
           let val (ty,insts) = instantiatePoly typ
            in (CONexp(dcon,insts),ty)
           end
       | INTexp (_,ty) => (OLL.push ty; (exp,ty))
       | WORDexp (_,ty) => (OLL.push ty; (exp,ty))
       | REALexp _ => (exp,realTy)
       | STRINGexp _ => (exp,stringTy)
       | CHARexp _ => (exp,charTy)
       | RECORDexp fields =>
           let fun h(l as LABEL{name,...},exp') = 
                    let val (nexp,nty) = expType(exp',occ,region)
                     in ((l,nexp),(l,nty))
                    end
               fun extract(LABEL{name,...},t) = (name,t)
               val (fields',tfields) = mapUnZip h fields
               val rty = map extract (sortFields tfields)
            in (RECORDexp fields',recordTy(rty))
           end
       | SELECTexp (l, e) =>
           let val (nexp, nty) = expType(e, occ, region)
               val res = mkMETAty ()
               val labtys = [(EU.labsym l, res)]
               val pt = VARty(mkTyvar(mkFLEX(labtys,infinity)))
            in (unifyTy(pt,nty); (SELECTexp(l, nexp), res))
               handle Unify(mode) =>
                 (err region COMPLAIN
                  (message("selecting a non-existing field from a record",mode))
                  (fn ppstrm =>
                   (PPType.resetPPType();
                    add_newline ppstrm;
                    add_string ppstrm "the field name: ";
                    (case l of LABEL{name,...} => ppSym ppstrm name);
                    add_newline ppstrm;
                    add_string ppstrm "the record type:    ";
                    ppType ppstrm nty; add_newline ppstrm;
                    add_string ppstrm "in expression:"; 
                    add_break ppstrm (1,2);
                    ppExp ppstrm (exp,!printDepth)));
                    (exp, WILDCARDty))
           end
       | VECTORexp(exps,_) =>
          (let val (exps',nty) = mapUnZip (fn e => expType(e,occ,region)) exps
               val vty = foldr (fn (a,b) => (unifyTy(a,b); b)) (mkMETAty()) nty
            in (VECTORexp(exps',vty), CONty(vectorTycon,[vty]))
           end handle Unify(mode) =>
	   (err region COMPLAIN
	     (message("vector expression type failure",mode))
             nullErrorBody; (exp,WILDCARDty)))
       | SEQexp exps => 
	   let fun scan nil = (nil,unitTy)
	         | scan [e] = 
                     let val (e',ety) = expType(e,occ,region)
                      in ([e'],ety)
                     end
		 | scan (e::rest) = 
                     let val (e',_) = expType(e,occ,region)
                         val (el',ety) = scan rest
                      in (e'::el',ety)
                     end
               val (exps',expty) = scan exps
            in (SEQexp exps',expty)
	   end
       | APPexp(rator, rand) =>
	   let val (rator',ratorTy) = expType(rator,occ,region)
	       val (rand',randTy) = expType(rand,occ,region)
               val exp' = APPexp(rator',rand')
	    in (exp',applyType(ratorTy,randTy))
	       handle Unify(mode) => 
	       let val ratorTy = prune ratorTy
		   val reducedRatorTy = headReduceType ratorTy
		in PPType.resetPPType();
		   if isArrowType(reducedRatorTy)
		   then (err region COMPLAIN
			  (message("operator and operand don't agree",mode))
			  (fn ppstrm =>
			   (add_newline ppstrm;
			    add_string ppstrm "operator domain: ";
			    ppType ppstrm (domain reducedRatorTy);
			    add_newline ppstrm;
			    add_string ppstrm "operand:         ";
			    ppType ppstrm randTy; add_newline ppstrm;
			    add_string ppstrm "in expression:";
			    add_break ppstrm (1,2);
			    ppExp ppstrm (exp,!printDepth)));
			 (exp,WILDCARDty))
		   else (err region COMPLAIN
			  (message("operator is not a function",mode))
			  (fn ppstrm =>
			    (add_newline ppstrm;
			     add_string ppstrm "operator: ";
			     ppType ppstrm (ratorTy); add_newline ppstrm;
			     add_string ppstrm "in expression:";
			     add_break ppstrm (1,2);
			     ppExp ppstrm (exp,!printDepth)));
			 (exp,WILDCARDty))
	       end
	   end
       | CONSTRAINTexp(e,ty) =>
	   let val (e',ety) = expType(e,occ,region)
	    in if unifyErr{ty1=ety,name1="expression", ty2=ty, name2="constraint",
			message="expression doesn't match constraint",
			region=region,kind=ppExp,kindname="expression",
			phrase=exp}
		then (CONSTRAINTexp(e',ty),ty)
		else (exp,WILDCARDty)
	   end
       | HANDLEexp(e,HANDLER h) =>
	   let val (e',ety) = expType(e,occ,region)
	       and (h',hty) = expType(h,occ,region)
               val exp' = HANDLEexp(e',HANDLER h')
	    in (unifyTy(hty, exnTy --> ety); (exp',ety))
	       handle Unify(mode) =>
		 (if unifyErr{ty1=domain(prune hty), name1="handler domain",
			     ty2=exnTy, name2="",
			     message="handler domain is not exn",
			     region=region,kind=ppExp,kindname="expression",
			     phrase=exp}
		     then unifyErr{ty1=ety, name1="body",
				   ty2=range(prune hty), name2="handler range",
				   message="expression and handler don't agree",
				   region=region,
				   kind=ppExp,kindname="expression",phrase=exp}
		     else false;
		  (exp,WILDCARDty))
	   end
       | RAISEexp(e,_) =>
	   let val (e',ety) = expType(e,occ,region)
               val newty = mkMETAty()
	    in unifyErr{ty1=ety, name1="raised", ty2=exnTy, name2="",
			message="argument of raise is not an exception",
			region=region,kind=ppExp,kindname="expression",phrase=exp};
	       (RAISEexp(e',newty),newty)
	   end
       | LETexp(d,e) => 
             if (!Control.CG.oldchecker) then
                 let val d' = decType0(d,LetDef(occ),region)
                     val (e',ety) = expType(e,occ,region)
                  in (LETexp(d',e'),ety) 
                 end
             else let val printAfterLet = (checkletpr1 (LETexp(d,e)) semiprob)
                      val d' = decType0(d,LetDef(occ),region)
                      val (e',ety) = expType(e,occ,region)
                   in 
                       (checkletpr2 printAfterLet semiprob;
                       (LETexp(d',e'),ety))
                   end

       | CASEexp(e,rules,isMatch) =>
	   let val (e',ety) = expType(e,occ,region)
	       val (rules',_,rty) = matchType(rules,occ,region)
               val exp' = CASEexp(e',rules', isMatch)
	    in (exp',applyType(rty,ety))
	       handle Unify(mode) => 
	       (if isMatch then
		    unifyErr{ty1=domain rty, name1="rule domain",
			     ty2=ety, name2="object",
			     message="case object and rules don't agree",
			     region=region,kind=ppExp,kindname="expression",phrase=exp}
                else 
                 let val decl = case rules 
                                 of (RULE(pat,_))::_ => 
				    VB{pat=pat,exp=exp,tyvars=ref[],boundtvs=[]}
                                  | _ => bug "unexpected rule list 456"
		  in unifyErr{ty1=domain rty, name1="pattern",
			      ty2=ety, name2="expression",
			      message="pattern and expression in val dec don't agree",
			      region=region,kind=ppVB,kindname="declaration",
			      phrase=decl}
                 end;
	        (exp,WILDCARDty))
	   end
		 (* this causes case to behave differently from let, i.e.
		    bound variables do not have generic types *)
       | FNexp(rules,_) => 
           let val (rules',ty,rty) = matchType(rules,occ,region)
            in (FNexp(rules',ty),rty)
           end
       | MARKexp(e,region) => 
           let val (e',et) = expType(e,occ,region)
            in (MARKexp(e',region),et)
           end
       | _ => bug "exptype -- bad expression"

and ruleType(RULE(pat,exp),occ,region) =  
    if (!Control.CG.oldchecker) then    
        let val occ = Abstr occ
            val (pat',pty) = patType(pat,lamdepth occ,region)
            val (exp',ety) = expType(exp,occ,region)
        in (RULE(pat',exp'),pty,pty --> ety)
        end
    else
        let val occ = Abstr occ
            val (pat',pty) = patType(pat,lamdepth occ,region)
            val printAfterRule = checkrulepr1 (RULE(pat,exp)) pty semiprob
            val oldrulebound = (!rulebound) 
            val _ = rbupdate pty
            val (exp',ety) = expType(exp,occ,region)
        in 
            (rulebound := oldrulebound);
            (checkrulepr2 printAfterRule semiprob);
            (RULE(pat',exp'),pty,pty --> ety)
        end

and matchType(l,occ,region) =
    case l
      of [] => bug "empty rule list in typecheck.matchType"
       | [rule] => 
	    let val (rule0,argt,rty) = ruleType(rule,occ,region)
	     in ([rule0],argt,rty)
	    end
       | rule::rest =>
	    let val (rule0,argt,rty) = ruleType(rule,occ,region)
		fun checkrule rule' =
		   let val (rule1,argt',rty') = ruleType(rule',occ,region)
		    in unifyErr{ty1=rty,ty2=rty', name1="earlier rule(s)",
				name2="this rule",
				message="types of rules don't agree",
				region=region,
				kind=ppRule,kindname="rule",phrase=rule'};
		       rule1
		   end
	     in (rule0::(map checkrule rest),argt,rty)
	    end

and decType0(decl,occ,region) : dec =
    let 
        val labels = Findvars.find_dec_vars(decl)  
        fun updateMonoConstraints labs =            
             let 
                 val _ = Semiunify.compress rulebound
                 val node = ref (!rulebound)  (* share these for all labs *)
             in  
                 monoConstraints := 
                 (map (fn lb => (lb,node)) labs) @ (!monoConstraints)
             end
    in
    case decl
      of VALdec vbs =>
	   let fun vbType(vb as VB{pat, exp, tyvars=(tv as (ref tyvars)), boundtvs}) =
	        let val (pat',pty) = patType(pat,infinity,region)
		    val (exp',ety) = expType(exp,occ,region)
                    val generalize = isValue exp (* orelse isVarTy ety *)
		 in unifyErr{ty1=pty,ty2=ety, name1="pattern", name2="expression",
			     message="pattern and expression in val dec don't agree",
			     region=region,kind=ppVB,kindname="declaration",
			     phrase=vb};
                   VB{pat=pat',exp=exp',tyvars=tv,
                      boundtvs=generalizePat(pat,tyvars,occ,generalize,region)}
                end
	    in (let val d = VALdec(map vbType vbs) 
                in  
                    if (!Control.CG.oldchecker) then d
                    else (updateMonoConstraints labels; d) 
                end)
	   end

       | VALRECdec(rvbs) =>
           if (!Control.CG.oldchecker) then    
 	   let val occ = Abstr occ

	       (* First go through and type-check all the patterns and
		  result-constraints, unifying with each other and with
		  the specified result type. *)

	       fun setType(rvb as RVB{var=VALvar{typ,...},exp,resultty,...}) = 
                   let val domainty = mkMETAtyBounded(lamdepth occ)
		       val rangety = mkMETAtyBounded(lamdepth occ)

		       val funty = domainty --> rangety

		       val _ = 
			   case resultty 
			     of NONE => true
			      | SOME ty =>
				 unifyErr{ty1=funty,ty2=ty,
					  name1="",name2="constraint",
					  message="type constraint of val rec dec\
					           \ is not a function type",
					  region=region,kind=ppRVB,
					  kindname="declaration", phrase=rvb}

		       fun f(FNexp(rules,_), region, funty) =
		             let fun unify a =
				  (unifyErr{ty1=a,name1="this clause",
				    ty2=funty,name2="previous clauses",
				    message="parameter or result constraints\
			                     \ of clauses don't agree",
					   region=region,kind=ppRVB,
					   kindname="declaration", phrase=rvb};
                                  ())
				       
				 fun approxRuleTy(RULE(pat,e)) =
				     let val (pat',pty) =
					     patType(pat,lamdepth occ,region)
				      in case e
					  of CONSTRAINTexp(e,ty) =>
					      (pat',pty-->ty,(e,region))
					   | e => (pat',pty-->rangety,(e,region))
				     end

				 val patTyExps = map approxRuleTy rules
				 val pats = map #1 patTyExps
				 val tys = map #2 patTyExps
				 val exps = map #3 patTyExps

				 fun doExp (e,region) =
				     let val (exp', ety) = expType(e,occ,region)
				      in unifyErr{ty1=ety, name1="expression",
					  ty2=rangety, name2="result type",
					  message="right-hand-side of clause\
					\ doesn't agree with function result type",
					  region=region,kind=ppRVB,
					  kindname="declaration",phrase=rvb};
					 exp'
				     end

                              in app unify tys;
				 typ := funty;
				 fn()=> 
				   FNexp(ListPair.map RULE (pats, map doExp exps),
						domain(prune(funty)))
			     end
		         | f(MARKexp(e,region),_,funty) = 
			     let val build = f(e,region,funty)
			      in fn()=> MARKexp(build(), region)
			     end
                         | f(CONSTRAINTexp(e,ty),region,funty) =
			     let val _ = 
				   unifyErr{ty1=ty, name1="this constraint",
					    ty2=funty, name2="outer constraints",
					    message="type constraints on val rec\
					             \ declaraction disagree",
					    region=region,kind=ppRVB,
					    kindname="declaration", phrase=rvb}
				 val build = f(e,region,funty)
			     in fn()=> CONSTRAINTexp(build(), ty)
			    end
			| f _ = bug "typecheck.823"
                   in f(exp,region,funty)		      
                  end

	      (* Second, go through and type-check the right-hand-side
	         expressions (function bodies) *)

	       fun rvbType(RVB{var=v,resultty,tyvars,boundtvs,...}, build) =
                      RVB{var=v,exp=build(),
                          resultty=resultty,tyvars=tyvars,boundtvs=boundtvs}
                  
	       val _ = debugmsg ">>decType0: VALRECdec"
               val builders = map setType rvbs
               val rvbs' = ListPair.map rvbType (rvbs,builders)
	    in 
               EU.recDecs rvbs'
	   end

           else 
               let val occ = Abstr occ

                   fun setType(RVB{var=VALvar{typ,...},resultty=NONE,...}) = 
                       typ := mkMETAtyBounded(1+lamdepth occ) (* was: mkRefMETAty(...) *)
                     | setType(rvb as RVB{var=VALvar{typ,...},resultty=SOME ty,...}) =
                       let 
                           val domainty = mkMETAtyBounded(lamdepth occ)
                           val rangety = mkMETAtyBounded(lamdepth occ)
                           val _ = unifyErr{ty1=domainty --> rangety,
                                            ty2=ty,
                                            name1="",name2="constraint",
                                            message="type constraint of val rec dec\
                                             \ is not a function type",
                                            region=region,kind=ppRVB,
                                            kindname="declaration", phrase=rvb}
                       in typ := ty end
                     | setType(_) = () 

                  (* Enforce user-constraints of rule-types (without unifying!) ...? *)

                   val oldrecbound = (!recbound)
                   val _ = recbound := revadd (map (fn (rvb as RVB{var=v,...}) => v) rvbs)
                                              oldrecbound
                   val thedeftys = ref([]:((ty ref * ty) list))
                       
                   fun rvbType(rvb as RVB{var=v as VALvar{typ,...},exp,
                                          resultty,tyvars,boundtvs}) =
                       let val (exp',ety) 
                           = (expType(exp,occ,region)) (* '93: Abstr(Rator occ) *)  
                       in 
                           (thedeftys := ((typ,ety)::(!thedeftys)));
                           RVB{var=v,exp=exp',resultty=resultty,tyvars=tyvars,boundtvs=boundtvs}
                       end
                     | rvbType _ = impossible "typecheck.786"
                       
                   val doclean = if (!topprob) then (topprob:= false; true) else false

                   val _ = (app setType rvbs)
                   val rvbs' = map rvbType rvbs

                   val _ = let fun f (tref,ty) = unifyTy(!tref,ty)
                           in (app f (!thedeftys)) end
                   val _ = recbound := oldrecbound
                       
                   val _ = updateMonoConstraints labels  
                   val _ = printSolvedIneqns := 
                           (checkrvbpr1 rvbs semiprob monoConstraints)

                   val _ = semiprob := 
                        (Semiunify.solve (!semiprob) (!monoConstraints)
                         handle Semiunify.SFAIL(mode,inequation,L) => 
                             (printSolvedIneqns := false;  
                              err region COMPLAIN(mode) 
                              (fn ppstrm =>
                               (PPType.resetPPType();
                                add_newline ppstrm;
                                add_string ppstrm "problem inequation";
                                add_newline ppstrm;
                                ppIneqlist ppstrm [inequation];
                                add_newline ppstrm;
                                add_string ppstrm "in current system";
                                add_newline ppstrm;
                                ppIneqlist ppstrm (!semiprob);
                                add_newline ppstrm;
                                add_string ppstrm "in declaration:";
                                add_break ppstrm (1,2);
                                PPAbsyn.ppDec (env,NONE) ppstrm 
                                (VALRECdec(rvbs),!printDepth)));
                              (L)))
                   val _ = checkrvbpr2 (!printSolvedIneqns) semiprob monoConstraints

                   val _ = if doclean then (semiprob := [];
                                            monoConstraints := [];
                                            eqcounter := 0; 
                                            PPType.resetPPType();
                                            topprob := true) 
                           else ()
               in 
                   EU.recDecs rvbs'
               end

       | EXCEPTIONdec(ebs) =>
	   let fun check(VARty(ref(UBOUND _))) = 
		     err region COMPLAIN
		         "type variable in top level exception type"
			 nullErrorBody
		 | check(CONty(_,args)) =
		     app check args
		 | check _ = ()
	       fun ebType(EBgen{etype=SOME ty,...}) = check ty
	         | ebType _ = ()
	       val _ = debugmsg ">>decType0: EXCEPTIONdec"
            in if TypesUtil.lamdepth occ < 1 then app ebType ebs else ();
               decl
	   end
       | LOCALdec(decIn,decOut) =>
	   let 
               val decIn' = decType0(decIn,LetDef occ,region)
               val decOut' = decType0(decOut,occ,region)
            in LOCALdec(decIn',decOut')
           end
       | SEQdec(decls) => 
           SEQdec(map (fn decl => decType0(decl,occ,region)) decls)
       | ABSTYPEdec{abstycs,withtycs,body} => 
	   let fun makeAbstract(GENtyc{stamp,arity,eq,path,kind}) = (eq := ABS)
		 | makeAbstract _ = bug "makeAbstract"
	       fun redefineEq(DATATYPEdec{datatycs,...}) =
		    (app (fn GENtyc{eq,...} => eq := DATA) datatycs;
		     EqTypes.defineEqProps(datatycs,nil,EntityEnv.empty))
	         | redefineEq(SEQdec decs) = app redefineEq decs
	         | redefineEq(LOCALdec(din,dout)) =
		    (redefineEq din; redefineEq dout)
	         | redefineEq(MARKdec(dec,_)) = redefineEq dec
	         | redefineEq _ = ()
	       val body'= decType0(body,occ,region)
	       val _ = debugmsg ">>decType0: ABSTYPEdec"
	    in app makeAbstract abstycs;
	       redefineEq body';
	       ABSTYPEdec{abstycs=abstycs,withtycs=withtycs,body=body'}
	   end
       | MARKdec(dec,region) => MARKdec(decType0(dec,occ,region),region)

      (* 
       * The next several declarations will never be seen ordinarily;
       * they are for re-typechecking after the instrumentation phase
       * of debugger or profiler. 
       *)
       | STRdec strbs => STRdec(map (strbType(occ,region)) strbs)
       | ABSdec strbs => ABSdec(map (strbType(occ,region)) strbs)
       | FCTdec fctbs => FCTdec(map (fctbType(occ,region)) fctbs)
       | _ => decl
      end

and fctbType (occ,region) (FCTB{fct,def,name}) =
      let fun fctexpType(FCTfct{param, argtycs, def}) =
  	        FCTfct{param=param, def=strexpType (occ,region) def,
	               argtycs=argtycs}
 	    | fctexpType(LETfct(dec,e)) =
	        LETfct(decType0(dec,LetDef occ,region), fctexpType e)
	    | fctexpType(MARKfct(f,region)) = MARKfct(fctexpType f,region)
            | fctexpType v = v
       in FCTB{fct=fct,def=fctexpType def,name=name}
      end

and strexpType (occ,region) (se as (APPstr{oper,arg,argtycs})) = se
  | strexpType (occ,region) (LETstr(dec,e)) =
      LETstr(decType0(dec,LetDef occ,region), strexpType (occ,region) e)
  | strexpType (occ,_) (MARKstr(e,region)) = 
      MARKstr(strexpType (occ,region) e, region)
  | strexpType _ v = v

and strbType (occ,region) (STRB{str,def,name}) =
    STRB{str=str,def=strexpType (occ,region) def,name=name}

val _ = debugmsg ">>decType: resetOverloaded"
val _ = resetOverloaded()
val _ = debugmsg ">>decType: OverloadedLit.reset"
val _ = OLL.reset ()

val _ = debugmsg ">>decType: calling decType0"
val dec' = decType0(dec, if toplev then Root else (LetDef Root), region);
val _ = debugPPDec("<<decType0 returning with: ", dec')

val _ = debugmsg ">>decType: OverloadedLit.resolve"
val _ = OLL.resolve ()
val _ = debugmsg ">>decType: resolveOverloaded"
val _ = resolveOverloaded env
val _ = debugmsg "<<decType: returning"
 in dec'
end (* function decType *)

val decType = Stats.doPhase (Stats.makePhase "Compiler 035 decType") decType

end (* local *)
end (* structure Typecheck *)

(*
 * $Log: typecheck.sml,v $
 * Revision 1.13  1997/11/11 05:29:50  dbm
 *   Fix error message (eliminate spurious newline).
 *
 * Revision 1.12  1997/11/07  05:43:48  dbm
 *   Improved wording or layout of some error messages.  Conditioned
 *   warning message regarding top-level value restriction violation
 *   on the flag Control.valueRestrictionTopWarn.
 *
 * Revision 1.11  1997/11/04  23:33:11  dbm
 *   Changed treatment of top level declarations that fail the value
 *   restriction.  Ungeneralized type variables are now instantiated
 *   to specially generated dummy types named X1, X2, etc.
 *
 * Revision 1.10  1997/10/08  21:12:28  dbm
 *   Fix for bug 1286 (Zhong).  The generalization procedure was complaining
 *   about the flex tyvars even if the depth of the tyvar was smaller than
 *   the current depth.
 *
 * Revision 1.9  1997/07/15  16:24:58  dbm
 *   Use new Control.valueRestrictionWarn flag to suppress local value
 *   restriction warnings.
 *
 * Revision 1.8  1997/04/16  22:26:35  dbm
 *   Fix for bug 317.  Recalculate equality props for datatypes in the
 *   body of an abstype after the type checker makes the abstract types
 *   abstract.
 *
 * Revision 1.7  1997/04/10  14:41:16  dbm
 *   Eliminated extra space from generalization error message.  Added local
 *   bind function and comment to generalizeTy.
 *
 * Revision 1.6  1997/04/02  04:22:36  dbm
 *   Fixed bug in 109.26 that improperly rejected some local declarations
 *    that failed the value restriction.
 *   Changed the treatment of FUNdecs to give priority to defining occurences
 *    of the defined function symbols over the applied occurences (which can
 *    sometimes occur before the defining occurences, causing confusing
 *    type error messages.
 *   Unified much of the type error message code via a function unifyErr.
 *
 * Revision 1.5  1997/03/27  17:19:16  dbm
 *   Changed test for top-level ungeneralized type variables to fix bug 1177.
 *
 * Revision 1.4  1997/03/22  18:28:01  dbm
 * Revised generalizeTy to fix bug 905/952.  Also changed the names of
 * the functions called for literal overloading.
 *
 * Revision 1.3  1997/03/18  16:46:19  dbm
 * Added test in META case of generalizeTy:gen to fix bugs 1066 and 274 (revised).
 *
 * Revision 1.2  1997/02/26  21:55:49  george
 *    Fix the bug, BUG 1139, on "unexpected type variables in mkPE" reported
 *    by Nikolaj (a.k.a. BUG 4 in his list of four bugs for 109.25)
 *
 *)

(* Modification over [Poly Rec] version 1.5 for SML/NJ Version 0.93:

 - replaced prune x = prune y by equalType(x,y) 
 - replaced compress by Semiunfiy.compressTy

 - replaced std_in by TextIO.stdIn
 - replaced input_line by TextIO.inputLine
 - replaced std_out by TextIO.stdOut
 - replaced flush_out by TextIO.flushOut

 - redefined polybound using eqVar, since Var is not an equality type

*)


